#/usr/bin/env python
# -*- coding: utf-8 -*-

#===============================================================================
# Author: Julien Eychenne
# License: GNU GPL
# Version: 05/06/2014
#===============================================================================

import codecs, copy
import numpy as np
np.random.seed()


#===============================================================================
# Some helper functions
#===============================================================================

def noise(sigma, kappa):
    '''Get a random number from the Gaussian distribution of mean 0 and variance sigma (= temperature)
    multiplied by the scaling factor kappa'''
    return np.random.normal(0,sigma) * kappa

def rand03():
    "Return a random number from a uniform distribution in [-0.3,0.3]"
    return (np.random.rand(1,1)[0][0]-0.5) * 3 / 5

def rand05():
    "Return a random number from a uniform distribution in [-0.5,0.5]"
    return np.random.rand(1,1)[0][0]-0.5

def assign_uniform_sonority(phones):
    sonority = { }
    for phone in phones:
        sonority[phone] = 1
    return sonority
        
def assign_random_sonority(phones):
    sonority = { }
    for phone in phones:
        sonority[phone] = np.random.rand(1,1)[0][0]*10
        
    return sonority
    
def simple_cost(expected, predicted):
    "Simple cost function where every mismatch costs 1"
    return sum(map(lambda x,y: int(x!=y), expected, predicted))

def phon2sonority(phon, sonority):
    "Get a sonority vector from a phonetic form"
    return np.array(map(lambda x: sonority[x], phon), dtype='float64')

def peaks2num(s):
    "Turn a string of L's, H's and O's into a numerical vector of 0's,1's and 2's respectively"
    s = s.replace('L','0').replace('O','1').replace('H','2')
    return map(int, s)

def get_weight_matrix(n, alpha, beta):
    '''Get tridiagonal matrix of weights in the form
       0 a 0 0 0
       b 0 a 0 0
       0 b 0 a 0
       0 0 b 0 a
       0 0 0 b 0
    '''
    W = np.zeros((n,n))
    for i in range(1,n):
        W[i-1, i] = alpha
        W[i, i-1] = beta
        
    return W
    
def check_value(param):
    "Alpha and beta must be in the interval [-0.5 0.5] to avoid non-convergence"
    return min(max(-0.5, param), 0.5)
    
def get_derived_sonority(input_son, alpha, beta):
    "Compute derived sonority in vectorized form, with |a*b| < 0.25"
    if abs(alpha*beta) > 0.25:
        print "WARNING:", input, "won't converge: skipped"
        return None
    
    n = len(input_son)
    W = get_weight_matrix(n, alpha, beta)
    I = np.eye(n)   
    
    # numpy won't straightforwardly transpose a row vector, so we work around it
    bias = np.array(input_son)[np.newaxis, :].T
    temp = np.dot(np.linalg.inv(I - W), bias)
    result = temp.T[0] # convert back to a row vector

    return result


def sonority2peaks(form):
    '''turn a sonority vector into
    a string like LHOLO, where:
    - L = local sonority minimum
    - H = local max
    - O = other
    See Larson (1993: 215)'''
    output = [ ]
    
    if len(form) == 1:
        return 'H'
    
    for i in range(len(form)):
        if i == 0:
            if form[i] < form[i+1]:
                output.append('L')
            else:
                output.append('H')
        
        elif i < len(form) - 1:
            if form[i-1] > form[i] < form[i+1]:
                output.append('L')
            elif form[i-1] < form[i] > form[i+1]:
                output.append('H')
            else:
                output.append('O')
        
        else:
            if form[i-1] < form[i]:
                output.append('H')
            elif form[i-1] > form[i]:
                output.append('L')
            else:
                output.append('O')
                
    return ''.join(output)

def load_lexicon(path, encoding = 'utf-8'):
    "Load a table file following the format full_phon\tphon\t\HLO\twhatever...\n"
    inf = codecs.open(path, 'r', encoding)
    lex = [ ]
    for ln in inf:
        ln = ln.strip()
        fields = ln.split('\t')
        lex.append(fields[:3])
    inf.close()
    
    return lex

#===============================================================================
# Model class: a simple structure to store the sonority values along with the parameters.
#===============================================================================

class DcnModel:
    def __init__(self, alpha, beta, sonority = { }):
        self.set_parameters(alpha, beta)
        self.sonority = sonority
        
    def set_parameters(self, alpha, beta):
        self.alpha = alpha
        self.beta = beta    

#===============================================================================
# Core functions
#===============================================================================

def simulated_annealing_training(lexicon, sonority,
         decay_rate = 0.995, kappa_ab = 0.2, kappa_son = 0.125, 
         temp_threshold = 0.001, temp = 1.0, trials = 1000):
    count = 0
    alpha = rand03(); beta  = rand03()
    temp_coeffs = [] 
    np.random.shuffle(lexicon)
    
    while count < trials and temp > temp_threshold:
        count += 1
        errors = 0

        for full_phon, input_phon, input_peaks in lexicon: 
            temp_coeffs.append(temp)
            input_son = phon2sonority(input_phon, sonority)
            derived_son = get_derived_sonority(input_son, alpha, beta)
            derived_peaks = sonority2peaks(derived_son)
            
            if input_peaks != derived_peaks:
                errors += 1
                input_vals = peaks2num(input_peaks)
                derived_vals = peaks2num(derived_peaks)
                
                for i in range(len(input_peaks)):
                    actual = input_vals[i]
                    predicted = derived_vals[i]
                    
                    if predicted > actual:
                        sonority[input_phon[i]] -= abs(noise(temp, kappa_son))

                    elif predicted < actual:
                        sonority[input_phon[i]] += abs(noise(temp, kappa_son))
                
                d_alpha = noise(temp, kappa_ab) # randomly modify alpha
                d_beta  = noise(temp, kappa_ab) # randomly modify beta   
                alpha += d_alpha #temp * np.random.normal(0,0.5)
                beta  += d_beta  #temp * np.random.normal(0,0.5)
                alpha = check_value(alpha)
                beta  = check_value(beta)
                new_temp = temp + np.sqrt(d_alpha**2 + d_beta**2)
                temp = min(new_temp, 1.0)
                
            # decay whether or not it was a success
            temp *= decay_rate
            if temp < temp_threshold:
                break
    converged = (count < trials)
    print "Epochs:", count
    perf = 100 - (float(errors)*100 / len(lexicon)) 
    print "Performance on training data:", perf   
    if not converged:
        print "WARNING: the temperature of the model is %f (threshold was %f) after %d trials" % (temp, temp_threshold, trials)
                
    return alpha, beta, temp_coeffs, converged, count, perf



# For discussion of kappa, see Larson(1993:227). 0.125 is used for sonority learning.
def train_simple(lexicon, sonority, decay_rate = 0.985, add_when_wrong = 0.15, 
                 temp_threshold = 0.001, temp = 1.0, trials = 100): 
    count = 0
    alpha = rand03(); beta  = rand03()
    temp_coeffs = [] 
    np.random.shuffle(lexicon)
    
    while count < trials and temp > temp_threshold:
        count += 1
        errors = 0

        for full_phon, input_phon, input_peaks in lexicon: 
            temp_coeffs.append(temp)
            input_son = phon2sonority(input_phon, sonority)
            derived_son = get_derived_sonority(input_son, alpha, beta)
            derived_peaks = sonority2peaks(derived_son)
            
            if input_peaks != derived_peaks:
                errors += 1
                for i in range(len(input_peaks)):
                    actual = input_peaks[i]
                    predicted = derived_peaks[i]
                    
                    if actual != predicted:
                        # sonority is too high
                        if predicted == 'H' or actual == 'L':
                            sonority[input_phon[i]] -= 0.1 
                        # sonority is too low
                        elif actual == 'H' or predicted == 'L':
                            sonority[input_phon[i]] += 0.1 
                
                alpha += temp * rand05()
                beta  += temp * rand05()
                alpha = check_value(alpha)
                beta  = check_value(beta)
                #new_temp = temp + np.sqrt(d_alpha**2 + d_beta**2)
                temp = min(temp+add_when_wrong, 1.0) # min(abs(annealing(temp, kappa_param)), 1.0)#min(1.0, new_temp)#
            # decay whether or not it was a success
            temp *= decay_rate
            if temp < temp_threshold:
                break
    converged = (count < trials)
    print "Epochs:", count
    print "Performance on training data:", 100 - (float(errors)*100 / len(lexicon))
    perf = 100 - (float(errors)*100 / len(lexicon)) 
    print "Performance on training data:", perf  
    if not converged:
        print "WARNING: the temperature of the model is %f (threshold was %f) after %d trials" % (temp, temp_threshold, trials)

    return alpha, beta, temp_coeffs, converged, count, perf

## Choose training method
#train = train_simple 
train = simulated_annealing_training
                
def report_training(alpha, beta, sonority):
    print "*" * 25
    print "Alpha:", alpha
    print "Beta:", beta
    print
    print "sonority = {"
    keys = sonority.keys()
    keys.sort()
    for k in keys:
        print "\t'%s': %f," % (k,sonority[k])
    print "}"
    
def test(lexicon, sonority, alpha, beta):
    "Run model on test data"
    good = []; bad = []

    
    for full_phon, input_phon, input_peaks in lexicon:
        input_son = phon2sonority(input_phon, sonority)
        derived_son = get_derived_sonority(input_son, alpha, beta)
        derived_peaks = sonority2peaks(derived_son)
        
        fields = [full_phon, input_phon, input_peaks, derived_peaks, input_son.tolist(), derived_son.tolist()]
        if input_peaks == derived_peaks:
            fields.append(1) # 1 in the last column means the model's prediction is correct  
            good.append(fields)
        else:
            fields.append(0) # 0 in the last column means the model's prediction is incorrect
            bad.append(fields)
    
    size = len(lexicon)
    success = float(size-len(bad))/ size * 100
    return good, bad, success
                
def report_test(bad, rate, alpha, beta, sonority = { }, path = u'./report/report_test.txt'):
    outf = codecs.open(path, 'w', 'utf-8')
    lines = []
    lines.append(u"*" * 25)
    lines.append("Alpha: %f" % alpha)
    lines.append("Beta: %f" % beta)
    lines.append("Success rate: %f" % rate) 
    lines.append(u"# of offenders: %d" % len(bad))  
    lines.append("*" * 25)
    lines.append("Offenders:")
    for b in bad:
        lines.append('\t'.join(map(lambda x: unicode(x), b)))

    if sonority:
        lines.append("*"*25)
        keys = sonority.keys()
        keys.sort()
        lines.append("sonority = {")
        for k in keys:
            lines.append("\t'%s': %f," % (k,sonority[k]))
        lines.append("}")
        
    for ln in lines:
        outf.write(ln+'\n')
      
    outf.close()
 
 
def run_ssp_baseline(lexicon, sonority):
    "Baseline according to the Sonority Sequencing Principle (alpha = 0; beta = 0)"

    result = test(lexicon, sonority, 0, 0)
    print "Baseline performance:", result[2]


def plot_decay(temperatures):
    import matplotlib.pyplot as plt
    plt.plot(temperatures)
    plt.show()
  
  
def run_once(training_lex, valid_lex, sonority, debug = False):
    a, b, temps, converged, count, perf = train(training_lex, sonority)
    result = test(valid_lex, sonority, a, b)
    print "Model performance:", result[2]
    if debug:
        report_training(a, b, sonority)
        plot_decay(temps)
        
      
def run_multiple(training_lex, valid_lex, sonority_hash, trials = 1, threshold = 0, debug = False):
    rates = [ ]
    train_perf = 0.0
    epochs = 0.0    
    # build an average model
    avg_son = { }
    avg_a = avg_b = 0
    
    for k in sonority_hash.keys():
        avg_son[k] = 0.0
    
    for i in range(1,trials+1):
        #sonority = copy.deepcopy(sonority_hash)
        sonority = assign_uniform_sonority(sonority_hash.keys())

        a,b,temps, converged, nepoch, tperf = train(training_lex, sonority)
        epochs += nepoch
        train_perf += tperf
        
        avg_a += a
        avg_b += b
        for k in sonority:
            avg_son[k] += sonority[k]
    
        good, bad, perf = test(valid_lex, sonority, a, b)
        
        if perf > threshold:
            threshold = perf
            if debug:
                report_training(a, b, sonority)
                plot_decay(temps) 

            report_test(bad, perf, a, b, sonority)

        rates.append(perf)    
        print "Trial %d: %f\tPerformance: %f" % (i, perf, np.mean(rates))
    
    for k in avg_son:
        avg_son[k] /= trials
    avg_a /= trials
    avg_b /= trials
    report_test([],0.0, avg_a, avg_b, avg_son, './report/avg_model_training.txt')
    train_perf /= trials
    epochs /= trials
        
    outf = open('./report/perf_test.log', 'w')
    for ln in map(str, rates):
        outf.write(ln+'\n')
    outf.close()    
    print "Average validation performance over %d trials: %f" % (trials, np.mean(rates)) 
    print "Minimum:", min(rates)
    print "Maximum:", max(rates)
    print "Average training performance: %f" % train_perf
    print "Average # of epochs:", epochs

def generate_model_report(model, lexicon, path):
    outf = codecs.open(path, 'w', 'utf-8')
    
    good, bad, perf = test(lexicon, model.sonority, model.alpha, model.beta)
    forms = good + bad
    forms.sort()
    
    for form in forms:
        outf.write('\t'.join(map(lambda x: unicode(x), form)) + '\n')
    outf.close()
    print "Model performance with alpha = %f / beta = %f: %f" % (model.alpha, model.beta, perf)
    
    
    
    
    
    
#===============================================================================
# Main            
#===============================================================================
if __name__ == '__main__':

    sonority = {
        'p': 1, 't':1, 'k':1,
        'b': 2, 'd':2, 'g':2,
        'f':3, 's':3, u'ʃ':3, 
        'v':4, 'z':4, u'ʒ':4,
        'm':5, 'n':5, u'ɲ':5, u'ŋ':5,
        'r':6, 'l':6,
        'w':7, u'ɥ':7, 'j':7,
        u'ɴ':7,  
        u'ə':8,
        'i':9, 'y':9, 'u':9,
        'e':11, u'ø':11, 'o':11, u'ɛ':11, u'œ':11, u'ɔ':11, '*':11, '6':11,
        'a':13       
    }
    
    
    
